# Imports from QT to handle signals and slots
from PyQt5.QtCore import QObject, pyqtProperty, pyqtSignal, pyqtSlot, QUrl

# Imports from Uranium and Cura to interact with the plugin system
from UM.Extension import Extension
from UM.Application import Application
from UM.PluginRegistry import PluginRegistry

# Imports from Uranium to enable internationalization
from UM.i18n import i18nCatalog
i18n_catalog = i18nCatalog("ContinuePrint")

# Imports from the python standard library to build the plugin functionality
from code import InteractiveInterpreter
import os
import platform
from io import StringIO
import sys
import html
from datetime import datetime
from UM.Scene.Iterator.DepthFirstIterator import DepthFirstIterator

#global variables
#On some windows systems cura has no permission to use the plugin directory for files
#us appdata on windows instead
if platform.system() == "Windows":
    dir_path = os.path.join(os.getenv('APPDATA'),"cura")
else:
    dir_path = os.path.dirname(os.path.realpath(__file__))



#This class is the extension and doubles as QObject to manage the qml
class PrintContinue(QObject, Extension, InteractiveInterpreter):
    #Create an api
    from cura.CuraApplication import CuraApplication
    api = CuraApplication.getInstance().getCuraAPI()

    # The QT signal, which signals an update for user information text
    userInfoTextChanged = pyqtSignal()

    # The constructor, which calls all the super-class-contructors, registers
    # our menu items and initializes the internal instance variables
    def __init__(self, parent = None) -> None:
        QObject.__init__(self, parent)
        Extension.__init__(self)
        InteractiveInterpreter.__init__(self, globals())

        #Create Menu entry
        self.setMenuName(i18n_catalog.i18nc("@item:inmenu", "ContinuePrint"))
        self.addMenuItem(i18n_catalog.i18nc("@item:inmenu", "Open ContinuePrint"), self.openContinueDialog)

        #Inizialise debug-log
        self.createLog()
        self.writeToLog("created log!")
        self.writeToLog("OS: "+platform.system())

        #Inzialize varables
        self.userText = ""
        self._continueDialog = None
        # here the height from which should be continued is saved
        self._printFromHeight = None
        # save here the dictionary containing the original GCode
        self._gcodeOriginal = None
        # save here the path of the original gcode
        self._gcodeOriginalPath = None


    #===Dialogue creation===============================================================================================

    # This method gets called he menu item for our plugin is clicked and
    # shows the print window if it has already been loaded
    def openContinueDialog(self):
        if self._continueDialog is None:
            self._continueDialog = self._createDialogue()
        self._continueDialog.show()
        self.writeToLog("Dialog opened")

    #====User Input=====================================================================================================

    #The QT property, which is computed on demand from our userInfoText when the appropriate signal is emitted
    @pyqtProperty(str, notify= userInfoTextChanged)
    def userInfoText(self):
        return self.userText

    #This method builds the dialog from the qml file and registers this class
    #as the manager variable
    def _createDialogue(self):
        qml_file_path = os.path.join(PluginRegistry.getInstance().getPluginPath(self.getPluginId()), "PrintContinue.qml")
        component_with_context = Application.getInstance().createQmlComponent(qml_file_path, {"manager": self})
        return component_with_context

    #is called when a key gets released in the height inputField (twice for some reason)
    @pyqtSlot(str)
    def heightEntered(self, text):
        #Is the textfield empty? Don't show a message then
        if text =="":
            self.writeToLog("Height-Textfield: Empty")
            self.userMessage("", "ok")
            self._printFromHeight = None
            return

        #Convert commas to points [+1 User Experience!]
        text = text.replace(",",".")

        self.writeToLog("Height-Textfield: read value "+text)

        #is the entered Text a number?
        try:
            float(text)
        except ValueError:
            self.userMessage("Entered height invalid: " + text,"bad")
            self.handleCommand(text)
            self._printFromHeight = None
            return
        self._printFromHeight = float(text)

        #Check if positive
        if self._printFromHeight <= 0:
            self.userMessage("Height must be positive. What do you expect?","bad")
            self._printFromHeight = None
            return

        self.writeToLog("Set printFromHeight to: " + text)

        #clear the message Field
        self.userMessage("", "ok")


    #Gets called when the Button <Generate> is clicked
    #Generates and loads the new gCode that continues the aborted print on the given height
    @pyqtSlot()
    def button1Pressed(self):
        self.writeToLog("---Button1 pressed")

        #get current scene object
        scene = Application.getInstance().getController().getScene()
        #If the scene does not have a gCode, don't generate the gCode
        if not hasattr(scene, "gcode_dict"):
            self.userMessage("Load a GCode file first.","bad")
            return
        gcode = getattr(scene, "gcode_dict")

        #when gcode == {} it means it a stl is loaded, but wasn't sliced yet
        if gcode == {}:
            #Slice for the user [+1 User Experience!]
            Application.getInstance().backend.forceSlice()
            self.setViewerAllVisible()
            self.userMessage("Slicing was needed. Try again when it's done!", "good")
            return

        #always generate new Code from the original gCode dict.
        gcode = self.getOriginalGCode(gcode)

        #When a GCode was directly loaded into cura, it gets saved in a list of lines.
        #It needs to be converted to the Layer-list format for better performance.
        gcode = self.convertToLayerList(gcode)

        #This checks if some Parameters exist and prints them to the log
        self.analyseGCode(gcode)

        #Is a hight value entered? If not, notify user and stop
        if self._printFromHeight is None:
            self.userMessage("Enter a height.", "bad")
            return
        height = self._printFromHeight

        #is the entered Height not to large? This can't be checked without the previous gCode operations.
        if height >= self.getObjectHeight(gcode):
            self.userMessage("The height " + str(height) + " is too large.", "bad")
            return

        #get the index of the Layer where we continue
        index = self.getContinueLayer(gcode, height)

        #generate an intern gCode file with the desired properties and load it
        self.generateContinueGCode(gcode, index)
        self.userMessage("Done. This is the GCode to continue with.","good")
        self.loadGeneratedGCode()

    #Gets called when the Button <Preview> is clicked
    #allows to quickly check if the measured height makes sense
    #by showing how the already existing part should look like
    @pyqtSlot()
    def button2Pressed(self):
        self.writeToLog("---Button2 pressed")

        #get current scene object
        scene = Application.getInstance().getController().getScene()
        #If the scene does not have a gCode, don't generate the gCode
        if not hasattr(scene, "gcode_dict"):
            self.userMessage("Load a gCode file first.", "bad")
            return
        gcode = getattr(scene, "gcode_dict")

        #when gcode == {} it means it a stl is loaded, but wasn't sliced yet
        if gcode == {}:
            #Slice for the user [+1 User Experience!]
            Application.getInstance().backend.forceSlice()
            self.setViewerAllVisible()
            self.userMessage("Slicing was needed. Try again when it's done!","good")
            return

        #Check if currently the original GCode is loaded, if not the preview makes no sense
        if not self.isGcodeOriginal(gcode):
            self.userMessage("Preview only works on the original GCode","bad")
            return

        #always generate new Code from the original gCode dict.
        gcode = self.getOriginalGCode(gcode)

        #When a GCode was directly loaded into cura, it gets saved in a list of lines.
        #It needs to be converted to the Layer-list format for better performance.
        gcode = self.convertToLayerList(gcode)

        #This checks if some Parameters exist and prints them to the log
        self.analyseGCode(gcode)
        # Is a hight value entered?
        if self._printFromHeight is None:
            self.userMessage("Enter a height.", "bad")
            return
        height = self._printFromHeight

        #is the entered Height not to large? This can't be checked without the previous gCode operations.
        if height >= self.getObjectHeight(gcode):
            self.userMessage("The height " + str(height) + " is too large.", "bad")
            return

        #get the index of the Layer where we continue
        index = self.getContinueLayer(gcode, height)

        #only show layers under the given index
        self.setViewerLayer(index-1)
        self.userMessage("Your printed part should look like this.","ok")

    #Gets called when the Button <Original> is clicked
    #Restores the original gCode. Doesn't work if .stl was original input, because there is no original code to load.
    @pyqtSlot()
    def button3Pressed(self):
        self.writeToLog("---Button3 pressed")

        #If loading original is not possible, notify user and stop
        if self._gcodeOriginalPath is None:
            self.userMessage("Error, try reimporting the gCode","bad")
            return

        #If original gCode was loaded, but looked different because the preview was active,
        #just reset the viewing layer instead of lading the gCode new. [+1 Performance!]
        scene = Application.getInstance().getController().getScene()
        gcode = getattr(scene, "gcode_dict")
        if self.isGcodeOriginal(gcode):
            self.setViewerAllVisible()
            self.writeToLog("Original: was already loaded")
            self.userMessage("loaded original code", "good")
            return

        #load the original gCode
        Application.getInstance()._openFile(self._gcodeOriginalPath)

        #reset Viewer
        self.setViewerAllVisible()

        self.userMessage("loading original code", "good")

    #Gets called when the circular <?>-Button is clicked
    #Some users we tested with, tried clicking the ? which would prevent the tooltip from showing So we tell them
    @pyqtSlot()
    def buttonHelpPressed(self):
        self.writeToLog("---ButtonHelp pressed")
        self.userMessage("Hover over the ? for a short guide","ok")

    #=====Text Output===================================================================================================

    #creates a new log file, and deletes the old one.
    def createLog(self):
        file = open(os.path.join(dir_path, "PrintContinueLog.txt"), "w")
        self.writeToLog("Creating Log in: " + str(dir_path))

    #writes the message to the log, includes timestamp, length is fixed
    def writeToLog(self, str):
        file = open(os.path.join(dir_path, "PrintContinueLog.txt"), "a")
        now = datetime.now()
        current_time = now.strftime("%H:%M:%S")
        file.write("["+current_time+"] ")
        #clip length
        if len(str) > 200:
            str = "+++To Long+++ "+ str[0:50]+"..."
        file.write(str)
        file.write("\n")

    #handles commands entered into the height textfield
    def handleCommand(self,text):
        #the command log opens the log file and shows its directory-path
        if text == "log":
            os.system("start " + os.path.join(dir_path, "PrintContinueLog.txt"))
            self.userMessage(str(dir_path),"ok")
            self.writeToLog("Command: log")
        #the command os shows the os name
        if text == "os":
            self.writeToLog("Command: os")
            self.userMessage(platform.system(), "ok")
        #the command dir lists all commands (yes, all three of them)
        if text == "dir":
            self.writeToLog("Command: dir")
            self.userMessage("Commands: log, os, dir", "ok")

    #Sends a user message to the Info Textfield, color depends on status (prioritized feedback)
    #Bad for Errors and Warnings
    #Grey for details and messages that aren't interesting for advanced users
    #Cura-Blue for successfully completing a task
    def userMessage(self, message, status):
        if status is "good":
            self.userText = "<font color='#151354'>" + message + "</font>"
        else:
            if status is "bad":
                self.userText = "<font color='#a00000'>" + message + "</font>"
            else:
                if status is "ok":
                    self.userText = "<font color='#9fa4b0'>" + message + "</font>"
                else:
                    self.writeToLog("Error: Invalid status: "+status)
                    return
        self.writeToLog("User Message: "+message)
        self.userInfoTextChanged.emit()

    #=====gCode Viewer==================================================================================================

    #set the layer until which the Model is shown in cura, and switch to "Preview"-Tab
    def setViewerLayer(self,value):
        #Make Sure the User sees whats going on [+1 User Experience!]
        Application.getInstance().getController().setActiveStage("PreviewStage")
        Application.getInstance().getController().setActiveView("SimulationView")
        simulationView = Application.getInstance().getPluginRegistry().getInstance().getPluginObject("SimulationView")
        simulationView.setLayer(value)
        self.writeToLog("Simulation View set to Layer"+str(value))

    #show the complete model
    def setViewerAllVisible(self):
        simulationView = Application.getInstance().getPluginRegistry().getInstance().getPluginObject("SimulationView")
        maxLayers = simulationView.getMaxLayers()
        self.setViewerLayer(maxLayers)

    #=====gCode Meta-Management=========================================================================================

    #loads the new generated gcode back into Cura
    def loadGeneratedGCode(self):
        filepath = os.path.join(dir_path, "generatedGCode.gcode")
        Application.getInstance()._openFile(filepath)
        #make code visible if preview was used before
        self.setViewerAllVisible()

    # get the Current loaded GCode File name
    def getPathOfCurrentGcode(self):
        for node in DepthFirstIterator(Application.getInstance().getController().getScene().getRoot()):
            gcode_filename = node.callDecoration("getGcodeFileName")
            if gcode_filename is not None:
                self.writeToLog("Gcode path found: "+str(gcode_filename))
                return gcode_filename
        self.writeToLog("Error: no currently loaded Gcode found")
        return None

    #returns the gcode dict of the original GCode (not the generated one, because layers are missing from it)
    def getOriginalGCode(self, currGcode):
        currGcodePath = self.getPathOfCurrentGcode()

        #If Code was Loaded from a .stl, no path will be found
        if currGcodePath is None:
            self._gcodeOriginal = currGcode.copy()
            self._gcodeOriginalPath = None
            self.writeToLog("First load from .stl")
            return currGcode

        #If no Gcode was loaded before, take the current code as the original
        if not self._gcodeOriginal:
            self._gcodeOriginal = currGcode.copy()
            self._gcodeOriginalPath = currGcodePath
            self.writeToLog("First Load of Gcode.")
            return currGcode

        #If Gcode was already loaded, check if it is a original or a generated one.
        #Always return the original code, not the changed one.
        if os.path.normpath(os.path.join(dir_path, "generatedGCode.gcode")) == os.path.normpath(currGcodePath):
            self.writeToLog("This is internal GCode")
            return self._gcodeOriginal.copy()
        else:
            self.writeToLog("This is external GCode")
            self._gcodeOriginal = currGcode.copy()
            self._gcodeOriginalPath = currGcodePath
            return currGcode

    #Checks if the given GCode is original
    def isGcodeOriginal(self,gcode):
        return gcode == self.getOriginalGCode(gcode)

    #=====gCode processing==============================================================================================

    #Returns a List, where all gCode commands for a layer are in one entry.
    #0 is header, 1 is Additional header, 2 is Layer 0
    def convertToLayerList(self, gcode):
        gcodeList = gcode.get(0)
        newLayerList = []

        #look if first Entry contains complete Header or just one line, to determine Format
        if len(str(gcodeList[0])) > 30:
            self.writeToLog("gCode seems to come from a .stl, already in Layer-List format")
            return gcode
        self.writeToLog("gCode seems to be loaded from a file, needs to get converted in Layer-List format")

        #Identify and deal with the special case when the Code is in Flavour: Marlin (zb for the printers from anycubic)
        #In this Case the printer has no Header-Tags, so we add them for our Plugin to work properly
        self.writeToLog(str(gcodeList[0]))
        if gcodeList[0] == ";FLAVOR:Marlin\n":
            self.writeToLog("GCode is in Flavour Marlin")
            newHeader = ";START_OF_HEADER\n"
            #put all commands before the engine-mention in the header
            for i in range(0,20):
                if str(gcodeList[i])[:32]==";Generated with Cura_SteamEngine":
                    gcodeList = gcodeList[i:]
                    break
                newHeader = newHeader + gcodeList[i]
            newHeader = newHeader + ";END_OF_HEADER\n"
            newLayerList.append(newHeader)

        #Convert the command list to the layer list
        layer = ""
        for x in gcodeList:
            layer = layer + x
            #x is last element of:
            #  Header                     normal layer                  additional Header            signature
            if x == ";END_OF_HEADER\n" or x[:14] == ";TIME_ELAPSED:" or x[:13] == ";LAYER_COUNT:" or x == ";End of Gcode\n":
                newLayerList.append(layer)
                layer = ""
        #Cora internally safes the gCode in a dict with the key 0, so we do that too to be polite
        gcode[0] = newLayerList
        self.writeToLog("Conversion complete!")
        return gcode


    #Analyses a given gCode Dictionary and prints results into the local log
    #When the gCode has a strange Format we see this here
    #gcode is a dictionary, it contains for key 0 a list with strings.
    #Each string in this list is the gCode for one layer. (Layer list format)
    def analyseGCode(self, gcode):
        #gets the list with layer instructions from the dictionary
        gcodeLayerList = gcode.get(0)

        #how many layers has the gCode depending on the List?
        layerCount = len(gcodeLayerList)-1

        self.writeToLog("===Start of gcode analysis====================")
        self.writeToLog("Layers: "+str(self.getLayerCount(gcode)))
        self.writeToLog("List length is " + str(len(gcode.get(0)))+ " (should be 4 larger than Layers on original gCode)")
        self.writeToLog("Object: Height" + str(self.getObjectHeight(gcode)))
        self.writeToLog("===End of gcode analysis======================")

    #returns the header of the given gCode Dictionary
    def gcodeHeader(self, gcode):
        #gets the list with layer commands from the dictionary
        gcodeLayerList = gcode.get(0)
        #The header is the first element (with index 0) in the list.
        gcodeHeader = str(gcodeLayerList[0])
        return gcodeHeader

    #returns the additional header of the given gCode Dictionary
    def gcodeAdditionalHeader(self, gcode):
        #gets the list with layer commands from the dictionary
        gcodeLayerList = gcode.get(0)
        #The AddHeader is the second element (with index 1) in the list.
        gcodeAddHeader = str(gcodeLayerList[1])
        return gcodeAddHeader

    #Returnes the String of gCode Commands for a given layer Index
    def gcodeLayer(self, gcode, layerNumber):
        #Index numbers 0 and 1 are additional information, layers start at index 2
        layerNumber = layerNumber+2
        #gets the list with layer instructions from the dictionary
        gcodeLayerList = gcode.get(0)
        gcodeLayer = str(gcodeLayerList[layerNumber])
        return str(gcodeLayer)

    #returns the Z value for the NEXT layer
    #layer is the gCode of the layer as a string
    def getNextLayerHeight(self, layer):
        #we look for a Z near the End. 300 is an empirical value
        indexOfZ = layer.find("Z",len(layer)-300)
        #we look for a line Break after the Z to find thr End of the number
        indexOfLineEnd = layer.find("\n",indexOfZ,indexOfZ+10)
        #the height should be between Z and \n
        height = layer[(indexOfZ)+1:indexOfLineEnd]
        #Check if we found a number
        try:
            float(height)
        except ValueError:
            self.writeToLog("Error: Searched layer Number. this is not a float: "+height)
            return
        return float(height)

    #extracts a float value from a given paragraph string, which stands behind the parameter string and
    #at the end of a line
    def extractGCodeInformation(self, paragraph, parameterStr):
        #we look for the parameter String
        IndexBehindParam = paragraph.find(parameterStr) + len(parameterStr)
        #we look for a line Break after the parameter to find the End of the number
        indexOfLineEnd = paragraph.find("\n", IndexBehindParam, IndexBehindParam + 20)
        #extract the value
        value = paragraph[IndexBehindParam:indexOfLineEnd]
        #Check if we found a number and return it as a float
        try:
            float(value)
        except ValueError:
            self.writeToLog("Error: For " + parameterStr+"this is not a float: " + value)
            return
        self.writeToLog("Found "+parameterStr+" with value " + value)
        return float(value)

    #returns the layer count given in the gCode (Attention: in some cases cura can't count correctly)
    def getLayerCount(self, gcode):
        return int(self.extractGCodeInformation(self.gcodeAdditionalHeader(gcode), "LAYER_COUNT:"))

    #returns the max height of the object given in the gCode
    def getObjectHeight(self, gcode):
        height =  self.extractGCodeInformation(self.gcodeHeader(gcode), "PRINT.SIZE.MAX.Z:")
        #some printers like those from anycubic feel special and use a different tag
        if height is None:
            height = self.extractGCodeInformation(self.gcodeHeader(gcode), ";MAXZ:")
        return height

    #Returnes the first Layer with a height larger than the given height
    def getContinueLayer(self, gcode, height):
        #Read layer Count from gCode
        layerCount = self.getLayerCount(gcode)

        #Search for Layer to continue on
        for i in range(1,layerCount-1):
            #There are some Meshes like "Doppelcone.stl" where the layer Count given in the Gcode is Wrong
            # -2 because the first Layer has list index 2 (header and Additional header are on 0 and 1)
            if i >= (len(gcode.get(0))-2):
                self.writeToLog("Error: tried to read from not existing Layer in getContinueLayer")
                return

            #getNextLayerHeight returns None when nothing valid was found
            lHeight = self.getNextLayerHeight(self.gcodeLayer(gcode, i))
            if lHeight is None:
                self.writeToLog("Error: No height information found in Layer "+str(i))
            else:
                #For continuing the layer should be above the already printed Part
                if lHeight > height:
                    self.writeToLog("Layer "+str(i+1)+" has height "+str(lHeight)+" and is the first one higher than "+ str(height))
                    return i+1 #Each layer contains height info for the NEXT layer so +1

        self.writeToLog("Error: this height was not found: "+str(height))

    #Adds important command to the first layer to be continued with
    # * some printer movements to prevent a visual bug in the gCode view [+1 User Experience!]
    # * the object cooling gets activated at the end of the first layer. Add that now. [+1 Printing Quality!]
    # * the Printer movement for the height to the layer, so that the layer is printed in the right height
    def addMissingLayerCommands(self, gcode, layerIndex):
        layer = self.gcodeLayer(gcode, layerIndex)
        height = self.getNextLayerHeight(self.gcodeLayer(gcode, layerIndex-1))
        maxHeight = self.getObjectHeight(gcode)

        #Find place in layer String where the new command should be added
        #(after the ";LAYER:X" comment, which is the first line)
        commandInsertIndex = layer.find("\n",6,11)
        if commandInsertIndex == -1:
            self.writeToLog("Error: previous height not found for layer "+str(layerIndex))
            return layer

        #Find the Coordinates where the new Layer starts (this Block is necessary
        #to prevent visual bugs in the Cura View)
        firstG1Index = layer.find("G1 X")
        if firstG1Index == -1:
            self.writeToLog("Error: first G1 command not found!" + str(layerIndex))
            return layer
        searchArea = layer[firstG1Index:firstG1Index+50]
        #this shoeld be a String like this: X34.564 Y12.53
        coordinatesXY = searchArea[searchArea.find("X"):searchArea.find(" E")]
        if coordinatesXY == "":
            self.writeToLog("Coordinates not found SeachArea: "+searchArea+" coordinates: "+coordinatesXY+"Indizes:"+str(searchArea.find("X"))+" "+str(searchArea.find(" E")))
        #This command moves the printer Head above the position to continue on
        startPosCommand = "G0 F500 "+coordinatesXY+" Z"+str(maxHeight)+"\n"

        #insert the new command                  move above XY   init Height                  activate cooling fan   close new commands
        command = "\n;start by ContinuePrint\n" +startPosCommand+"G0 F500 Z"+str(height)+"\n"+"M106 S255\n"  +       ";end by ContinuePrint\n"
        self.writeToLog("Added command to layer "+str(layerIndex)+": \n"+command)
        return layer[:commandInsertIndex] + command + layer[commandInsertIndex+1:]


    #generates the new gCode by using Header, Additional Header and all layers from the given index on
    #Saves this to a file "generatedGCode.gcode"
    def generateContinueGCode(self, gcode, layerIndex):
        #make new list
        newLayerList = []

        #create new gCode file
        open(os.path.join(dir_path, "generatedGCode.gcode"), "w")
        self.writeToLog("Gcode File: generated")

        #add Header
        file = open(os.path.join(dir_path, "generatedGCode.gcode"), "a")
        file.write(self.gcodeHeader(gcode))
        newLayerList.append(self.gcodeHeader(gcode))
        #add Additional Header
        file.write(self.gcodeAdditionalHeader(gcode))
        newLayerList.append(self.gcodeAdditionalHeader(gcode))
        self.writeToLog("Gcode File: Added header and Additional header")

        #add first new layer, with inserted height commands and fan speed
        firstLayer = self.addMissingLayerCommands(gcode, layerIndex)
        file.write(firstLayer)
        newLayerList.append(firstLayer)

        #add all the missing layers which need to get printed 8not the first) and final printer movements
        layerCount = len(gcode.get(0))-2
        for i in range(layerIndex+1,layerCount):
            file.write(self.gcodeLayer(gcode,i))
            newLayerList.append(self.gcodeLayer(gcode,i))
        self.writeToLog("Gcode File: Added missing layers and final printer movements")
        gcode[0] = newLayerList
        return gcode
